# This file gathers functions useful for the implementation : 
# 'basisfunc()'             : Generates matrix of regressors.
# 'simdata()'               : Generates data according to location, loc.sc. and DR loc.scale DGPs.
# 'califunc()'              : Calibrates the parameters according to LOC, DR or QR models.
# 'crossvalidate()'         : Cross validates QR model for ASF.

# This version: June 2019.


# ==========================================
#
#  I - Generates matrix of regressors	  	   	
#
# ==========================================

basisfunc <- function(X,R=NULL,qVhat,Vpolys=1){
  
  ifelse(ncol(as.matrix(X))==1, X <- cbind(X,R,I(X*R)), X <- cbind(X,R) )
  x      <- NULL
  for(k in 1:Vpolys){ 
        x    <- cbind(x,qVhat^k, I(qVhat^k * X))
  }
  Xmat   	 <- cbind(X, x); 
  
  return(Xmat)
}


# ==========================================
#
#  II - FUNCTION TO SIMULATE DATA	  	   	
#
# ==========================================

simdata <- function(avec, bvec, n = 100, rho = .5, sde = 1, sdv = 1, Z = NULL, seed = 999, dgp = "loc", gridY = NULL, x0 = NULL){;
# Function to simulate a lin triangular model DGP.
# Admits only scalar X. Takes Z as given
# Also returns the value of the DSF ofor a grid of Y values and value x0 of X.

    Dat    <- NULL;
    Dat$U  <- rnorm(n);
  	Dat$V  <- rnorm(n);	
    Dat$Z  <- Z

    if(dgp == "loc"){
              Dat$X   <- avec[1:2] %*% rbind(1,Dat$Z) + (avec[3:4] %*%rbind(1,Dat$Z))*sdv*Dat$V;
              Dat$Y   <- bvec[1:2] %*% rbind(1,Dat$X) + (bvec[3:4] %*%rbind(1,Dat$X))*sde*(rho*Dat$V +  sqrt(1-rho^2) * Dat$U);
              if(length(x0)>0){
                ifelse(length(x0)==1, Dat$U0  <- (gridY - rep(bvec[1:2] %*% rbind(1,x0),length(gridY)))/(sde*(bvec[3:4] %*%rbind(1,x0))), Dat$U0 <- "NA") 
                Dat$mux <- bvec[1:2] %*% rbind(1,x0)
              }
    }
    
    if(dgp == "qr"){
      Dat$X   <- avec[1:2] %*% rbind(1,Dat$Z) + (avec[3:4] %*%rbind(1,Dat$Z))*Dat$V;
      Dat$Y   <- bvec[1:2] %*% rbind(1,Dat$X) + (bvec[3:4] %*%rbind(1,Dat$X))*(rho*Dat$V +  sqrt(1-rho^2) * Dat$U);
      if(length(x0)>0){
        ifelse(length(x0)==1, Dat$U0  <- (gridY - rep(bvec[1:2] %*% rbind(1,x0),length(gridY)))/(sde*(bvec[3:4] %*%rbind(1,x0))), Dat$U0 <- "NA") 
        Dat$mux <- bvec[1:2] %*% rbind(1,x0)
      }
    }
    
  	if(dgp == "dr"){
          	  Dat$X   <- (Dat$V - (avec[1:2] %*% rbind(1,Dat$Z))) / (avec[3:4] %*%rbind(1,Dat$Z));
          	  Dat$Y   <- ((rho*Dat$V + ( (1-rho^2)^.5 ) * Dat$U) - (bvec[1:2] %*% rbind(1,Dat$X))) / (bvec[3:4] %*%rbind(1,Dat$X))
          	  if(length(x0)>0){
          	    ifelse(length(x0)==1, Dat$U0  <- (bvec[c(1,3)] %*% rbind(1,gridY)) + (bvec[c(2,4)] %*%rbind(1,gridY))*x0, Dat$U0 <- "NA")
          	    Dat$mux <- -(bvec[1:2] %*% rbind(1,x0)) / (bvec[3:4] %*%rbind(1,x0))
          	  }
  	}
    
  	ifelse(length(x0)>0, ans <- list(Y = Dat$Y, X = Dat$X, Z = Dat$Z, V = Dat$V, U = Dat$U, U0 = as.numeric(Dat$U0), mux = Dat$mux ),
  	       ans <- list(Y = Dat$Y, X = Dat$X, Z = Dat$Z, V = Dat$V, U = Dat$U ))
	
  	return(ans)
  	}
		


# =========================================
#    									                     	
#			III - Calibration                     
#										                       
# =========================================

califunc <- function(Y, X, Z, qY = NULL, taus1, taus2, trim = NULL, dgp = "loc", link = "probit"){;

  if(dgp == "loc"){
    
    # Calibration First Stage
    coefFS       <- coef(lm(X~Z))
    inter1coef   <- c(coefFS[1],1);      # Calibrated coefs for intercept
    Zcoef	    	 <- c(coefFS[2],0);      # Calibrated coefs for Z
    eta          <- X - fitted(lm(X~Z))
    resid        <- eta/sd(eta)
    ecdf         <- (rank(resid)-.5)/length(X)
    sdv          <- sd(eta)
    qVhat        <- qnorm(ecdf)
    
    # Calibration Second Stage
    coefSS       <- lm(Y ~ X + qVhat)$coef
    inter2coef   <- c(coefSS[1],1);      # Calibrated coefs for intercept
    Xcoef	    	 <- c(coefSS[2],0);      # Calibrated coefs for Z
    eps          <- Y - (coefSS[1]+coefSS[2]*X) 
    sde          <- sd(eps)
    resid        <- eps/sde
    ecdf         <- (rank(resid)-.5)/length(X)
    qeps         <- qnorm(ecdf)

    rho          <- cor(qVhat,qeps)
  }
  
  if(dgp == "qr"){
    
    # Calibration First Stage
    FS           <- CF(X = X, Z = Z, taus = taus1, FS = "QR");        		# First Stage
    Vhat       	 <- FS$CV; 
    coefsFS      <- FS$coefs;          	
    
    inter1coef   <- coef(lm(coefsFS[,1] ~ qnorm(taus1)));    # Calibrated coefs for intercept
    Zcoef	    	 <- coef(lm(coefsFS[,2] ~ qnorm(taus1)));    # Calibrated coefs for Z
    qVhat        <- qnorm(Vhat)
    
    # Build trimmed sample if needed
    if(trim[1]>0){
      iout         <- c(which(X<quantile(X,trim[1])),which(X>quantile(X,trim[2])))
      Y            <- Y[-iout]
      X            <- X[-iout]
      qVhat        <- qVhat[-iout]
    }
    
    # Calibration Second Stage
    coefs        <- t(coef(rq(Y ~ cbind(X, qVhat, I(X*qVhat)), tau = taus2))) 
    res.dsf      <- strucfuncs(Y = Y, X = X, Z = Z, taus1 = taus1, taus2 = taus2, trim = trim, qY = qY, Ybar = Y,  gridX = X, FS = "QR", SS = "QR")$DSF;
    qeps         <- qnorm(diag(res.dsf))
    rho          <- cor(qVhat,qeps)
    rho_fac      <- (1-(rho^2))^(.5)

    inter2coef   <- coef(lm(coefs[,1] ~ I(rho_fac * qnorm(taus2)) ) );      # Calibrated coefs for intercept
    Xcoef	   	   <- coef(lm(coefs[,2] ~ I(rho_fac * qnorm(taus2)) ) );      # Calibrated coefs for X
    
    sdv          <- NULL
    sde          <- NULL
    
  }
  
  if(dgp == "dr"){
    
    # Calibration First Stage
    FS           <- CF(X = X, Z = Z, taus = taus1, FS = "DR", link = link);        		# First Stage
    qVhat 	  	 <- qnorm(FS$CV); 
    coefsFS      <- FS$coefs0;          	
    qX           <- FS$grid

    inter1coef   <- coef(lm(coefsFS[1,] ~ qX));    # Calibrated coefs for intercept
    Zcoef	    	 <- coef(lm(coefsFS[2,] ~ qX));      # Calibrated coefs for Z

    # Build trimmed sample if needed
    if(trim[1]>0){
      iout         <- c(which(X<quantile(X,trim[1])),which(X>quantile(X,trim[2])))
      Y            <- Y[-iout]
      X            <- X[-iout]
      qVhat        <- qVhat[-iout]
    }
  
    # Calibration Second Stage
    gridY         <- quantile(Y,qY)     # Grid of Y values for distribution regression estimator
    coefs         <- distrfun(Y = Y, X = cbind(X, qVhat), gridY = gridY, link = link)$coefs0
    res.dsf       <- strucfuncs(Y = Y, X = X, Z = Z, taus1 = taus1, taus2 = taus2, trim = trim, qY = qY, Ybar = Y, gridX = X, FS = "DR", SS = "DR", link = link)$DSF;
    qeps          <- qnorm(diag(res.dsf))
    rho           <- cor(qVhat,qeps)
    rho_fac       <- (1-rho^2)^(.5)

    inter2coef    <- coef( lm(coefs[1,] ~ 0 + rep(rho_fac^(-1),length(gridY)) + I(rho_fac^(-1) * gridY) ) );      # Calibrated coefs for intercept
    Xcoef	   	    <- coef( lm(coefs[2,] ~ 0 + rep(rho_fac^(-1),length(gridY)) + I(rho_fac^(-1) * gridY) ) );      # Calibrated coefs for X
    
    sdv           <- NULL
    sde           <- NULL
  }
  
  
  # Store calibrated coefs
  avec <- c( inter1coef[1], Zcoef[1], inter1coef[2], Zcoef[2] ) 
  bvec <- c( inter2coef[1], Xcoef[1], inter2coef[2], Xcoef[2] ) 
  
  ans <- list(avec = avec, bvec = bvec, rho = rho, sde = sde,  sdv = sdv)

  return(ans)	
  
}


# ==========================================
# 
#  IV - CROSS VALIDATION						                 
#
# ==========================================

crossvalidate <- function(Y, X, Z, R = NULL, Vpolys, taus,  FS = "QR"){
        
        fit    <- numeric(length(X))
        
        # First Stage
        FS   	 <- CF(X = X, Z = Z, R = R, taus = taus, FS = FS);        		
        qVhat  <- qnorm(FS$CV)
        x      <- basisfunc(X=X,R=R,qVhat=qVhat,Vpolys=Vpolys)
        # Second Stage  
        #for(i in 1:length(X)){
        #      formula2 	    <- Y[-i] ~ x[-i,];
        #      fit2          <- lm(formula = formula2);
        #      fit[i]        <- t(coef(fit2)) %*% rbind(1,as.matrix(x[i,]));
        #}
        
        #SSE   <- sum((Y-fit)^2)
        mod   <- lm(Y~x)
        err   <- mod$res
        denom <- 1-lm.influence(mod)$hat
        SSE   <- mean((err/denom)^2)
        
        return(SSE)
        }

