# This is the main functions for the estimation of structural functions in the manuscript.

# Includes:
# 'CF()'                    : Control function estimation
# 'strucfuncs()'            : Wrapper function for structural functions of interest.
# 'qdsf()'                  : QSF and DSF estimation.
# 'asf()'                   : ASF estimation.
# 'cdfv()'                  : F(Y|x,R,V) estimation. For fixed X=x. Used in 'qdsf()'.

# This version : June 2019


# =========================================
#    								       	
#	  I - CONTROL FUNCTION
# 										   
# =========================================

CF <- function(X, Z, R = NULL, taus = NULL, FS = "QR", link = "logit", weights = NULL){

  # Function to estimate the CDF of X given Z at n sample points
  # Includes quantile, distribution and mean regression methods.
  # FS = "QR"    : quantile regression.
  # FS = "DR"    : distribution regression.
  # FS = "LS"    : LS residuals.
  
  Dat			     	<- NULL;
  Z             <- cbind(Z,R)
  formula		  	<- X ~ Z;
  nobs          <- length(X);
  ntaus         <- length(taus)
  
  if(FS == "QR"){
    eps       <- min(taus1) 
    fit	  	  <- rq(formula = formula, tau = taus, weights = weights);
    coefs  	  <- t(coef(fit));
    u	        <- fit$tau;
    Dat$cqf   <- coefs %*% rbind(1, t(Z));      
    Dat$cdf   <- sapply(1:nobs, function(i) ifelse(sum(Dat$cqf[ ,i] <= X[i])  !=0, u[sum(Dat$cqf[ ,i] <= X[i])], eps))
  };
  
  if(FS == "DR"){
    gridX     <- quantile(X,seq(0,1,length.out=length(taus)))
    fit       <- distrfun(Y = X, X = Z, gridY = gridX, link = link, weights = weights, smplval = TRUE)
    coefs0    <- fit$coefs0
    coefs     <- fit$coefs   
    Dat$cdf   <- fit$cdfn
  }
  
  if(FS == "LS"){
    fit  	    <- lm(formula = formula, weights = weights);
    coefs     <- fit$coef
    res       <- X - fit$fitted
    resid     <- res/sd(res)
    Dat$cdf   <- (rank(resid)-.5)/nobs
  }
  
  if(FS == "DR"){ ans <- list(CV = Dat$cdf, coefs = coefs, coefs0 = coefs0, gridX = gridX)	}
  if(FS != "DR"){ ans <- list(CV = Dat$cdf, coefs = coefs) }
  
  return(ans)
}


# ========================================
#    								       	
#	  II - STRUCTURAL FUNCTIONS     
#   Wrap-up function for asf, dsf and qsf.
#   Only scalar X and R.
#
# ========================================


strucfuncs <- function(Y, X, XSS = NULL, Z = NULL, R = NULL, qVhat = NULL, trim = c(0,1), integration = "Smplave", tausqsf = c(.25,.5,.75), taus1 = NULL, taus2, qY, qYbar=NULL, Ybar = NULL, gridX,
                       FS = "QR", SS = "QR", link = "logit", rSS = 0, Vpolys = 1, weights = NULL, 
                       methodASF = "approx", lb = 0, ub = 1, subdivisions = 100, splinefuns = NULL, QDSF = TRUE){;

     simus <- NULL;if(length(qVhat)>0){ simus <- TRUE }     # Flag: is it a MC simu?

     # First Stage
     if(length(qVhat)==0){
        FS   	        <- CF(X = X, Z = Z, R = R, taus = taus1, link = link, FS = FS, weights = weights);        	
  	    qVhat 	  	  <- qnorm(FS$CV); 
     }
  
     # Build trimmed sample if needed
     if(trim[1]>0){
        iout         <- c(which(X<quantile(X,trim[1])),which(X>quantile(X,trim[2])))
        Y            <- Y[-iout]
        X            <- X[-iout]
        XSS          <- XSS[-iout]
        R            <- R[-iout] 
        qVhat        <- qVhat[-iout]
        weights      <- weights[-iout]
     }
     if(length(XSS)==0){ XSS <- X }
     
     # AQDSF
     ifelse(min(qY)>0, Yt <- Y[-c(which(Y<quantile(Y,min(qY))),which(Y>quantile(Y,max(qY))))], Yt <- Y )  # For second stage with trimmed Y
     ifelse(length(qY)>=length(Y), gridY <- Yt, gridY <- quantile(Y,qY))                     # Grid of Y values for DSF approximation
     YS            <- seq(min(gridY),max(gridY),length.out=length(gridY))                    # Grid of Y values for DSF integration
     if(length(Ybar)==0){ ifelse(length(qYbar)==0, Ybar <- YS, Ybar <- quantile(Y,qYbar)) }  # Grid of Y values for DSF estimation

     if(length(QDSF)>0){
     resQDSF       <- qdsf(Y = Y, X = XSS, R = R, qVhat = qVhat, integration = integration, tausqsf = tausqsf, taus2 = taus2, YS = YS, Ybar = Ybar,
  	                      gridY = gridY, gridX = gridX, link = link, SS = SS, rSS = rSS, Vpolys = Vpolys, weights = weights, splinefuns = splinefuns);
     }
     resASF        <- asf(Y = Y, X = XSS, R = R, qVhat = qVhat, gridX = gridX, YS = YS, resQDSF = resQDSF, methodASF =  methodASF, lb = lb, ub = ub,
                          subdivisions = subdivisions, SS = SS, Vpolys = Vpolys, weights = weights)
        
  	 #  Objects to return				   
     if(length(QDSF)>0){
       if(length(weights)==0 && length(simus)==0){ ans <- list(ASF = resASF, QSF = resQDSF$QSF, DSF = resQDSF$DSF,
                                                              gridX = gridX, gridY = gridY, YS = YS, coefs2 = resQDSF$coefs, taus1 = taus1, taus2 = taus2, qVhat = qVhat) }
       if(length(weights)>0){ ans <- list(ASF = resASF, QSF = resQDSF$QSF, DSF = resQDSF$DSF) } # Light load for bootstrap and simulations  	  
       }
     if(length(QDSF)==0){
       if(length(weights)==0 && length(simus)==0){ ans <- list(ASF = resASF, gridX = gridX, gridY = gridY, YS = YS, taus1 = taus1, taus2 = taus2, qVhat = qVhat) }
       if(length(weights)>0){ ans <- list(ASF = resASF) } # Light load for bootstrap and simulations  	  
     }
     if(length(simus)>0){   ans <- list(ASF = resASF, QSF = resQDSF$QSF) } # Light load for bootstrap and simulations  	  
     
     return(ans);
  	  
     };



# =========================================
#    								       	
#	  III - QSF and DSF    
# 										   
# =========================================

qdsf <- function(Y, X, R = NULL, qVhat = NULL, gridX, taus2 = NULL, tausqsf = seq(.25,.75,by=.25), gridY, YS = NULL, Ybar = NULL,
                 link = "logit", SS = "QR", rSS = 0, integration = "Smplave", Vpolys = 1, weights = NULL, splinefuns = NULL){
  
  # Computes the Quantile and Distribution Structural Functions of Imbens and Newey (2009)
  # rSS == 0  : Rearranged distribution regression for the second stage.
  
  QSF			<- NULL;
  DSF			<- NULL;
  DSFint	<- NULL;
  nY      <- length(gridY)		
  nobs    <- length(Y)
  x       <- basisfunc(X=X,R=R,qVhat=qVhat,Vpolys=Vpolys)
    
  # Second Stage: get regression coefficients
  if(SS == "QR"){ coefs  <- t(coef(rq(Y ~ x, tau = taus2, weights = weights))) }
  if(SS == "DR"){ coefs  <- distrfun(Y = Y, X = x, gridY = gridY, link = link, weights = weights)$coefs0 }

  # Third Stage: get structural functions
  if(integration == "simu"){ 
    #nint.grid <- length( taus2 )
    nint.grid <- nobs
    qVhat     <- runif( nint.grid )
    nobs      <- nint.grid
  }
  delta  <- YS[2] - YS[1]
  indicY <- 1*(YS>0)
  
  for (j in 1:length(gridX)){
    
      cdfvmat  <- cdfv(v=qVhat,x=gridX[j],R=R,gridY=gridY,coefs=coefs,taus2=taus2,SS=SS,Vpolys=Vpolys,nobs=nobs,integration=integration,link=link,rSS=rSS,splinefuns=splinefuns)
      if(length(weights)>0){ cdfvmat <- matrix(weights,nr=nrow(cdfvmat),nc=ncol(cdfvmat),byrow=T)*cdfvmat }
      
      #DSF
      ifelse(length(weights)==0, dsf <- apply(cdfvmat, 1, mean), dsf <- apply(cdfvmat, 1, sum)/sum(weights) )
      if(SS=="DR" && rSS == 0){ dsf <- rearrangement(as.data.frame(gridY), dsf) }
      dsfunc   <- approxfun(gridY, dsf, rule = 2)
      DSF      <- cbind(DSF, dsfunc(Ybar))
      
      # QSF
      DSFint   <- cbind(DSFint, dsfunc(YS))
      indicDSF <- NULL
      for(i in 1:length(tausqsf)){ indicDSF <- cbind(indicDSF, 1*(DSFint[,j] >= tausqsf[i])) } 
      QSFx     <- delta * apply(indicY - indicDSF,2,sum)
      
      QSF	     <- rbind(QSF, QSFx)
      
    }
  
  
  ans <- list(QSF=QSF, DSF = DSF, DSFint = DSFint, coefs = coefs, gridX=gridX) 
  
  return(ans)
  
};



# =========================================
#    								       	
#	  IV - ASF    
# 										   
# =========================================

asf <- function(Y, X, R = NULL, qVhat = NULL, gridX, YS, resQDSF = NULL, methodASF = "approx", lb = 0, ub = 1, subdivisions = 100, SS = "QR", Vpolys = 1, weights = NULL){
  
  # Computes the Average Structural Function
  # Requires computation of the DSFs first, exept for methodASF == "LS" which implements OLS.
  # methodASF == "intQSF": ASF is computed by averaging QSFs over probability levels. QSFs must be provided.

  ngridX <- length(gridX)
  
  if(methodASF == "approx"){
      delta <- YS[2] - YS[1]
      indic <- 1*(YS>0)
      ASF   <- delta * apply(indic - resQDSF$DSFint,2,sum)
  }
  
  if(methodASF == "intQSF"){ ASF <- apply(resQDSF$QSF, 1, mean, na.rm = TRUE) }
  
  if(methodASF == "LS"){
    x       <- basisfunc(X = X, R = R, qVhat = qVhat, Vpolys = Vpolys)
    coefASF <- lm(Y~x, weights = weights)$coef
    
    ifelse(length(R)==0, ASF <- coefASF[1] + coefASF[2] * gridX, ASF <- coefASF[1] + coefASF[2] * gridX + coefASF[3] * mean(R) + coefASF[4] * mean(R) * gridX )
  }
  
  
  return(ASF)
}


# =========================================
#    								       	
#	  V - F(Y|x,R,V)    
# 										   
# =========================================

cdfv <- function(v,x,R=NULL,gridY,coefs,taus2,SS="QR",Vpolys=1,nobs,integration="Smplave",link="probit",rSS=0,splinefuns=NULL){
  
  nY      <- length(gridY)
  nv      <- length(v)
  nspl    <- length(splinefuns)
  if(nspl==0){
    x_iv    <- rbind(1,t(basisfunc(X=rep(x,nv),R=R,qVhat=v,Vpolys=Vpolys)))
  }
  if(length(splinefuns)>0){
    splines_x <- NULL
    for(l in 1:nspl){ splines_x <- c(splines_x, splinefuns[[l]](x)) }
    x_iv    <- rbind(1,t(basisfunc(X=t(kronecker(as.matrix(splines_x), matrix(1, ncol = nv, nrow = 1))),R=R,qVhat=v,Vpolys=Vpolys)))
  }
  cdfvmat <- matrix(0, ncol = nv, nrow = nY);
  
  if(SS == "QR"){
    cqfv <- coefs %*% x_iv 
    #time1 <- system.time(
    for (i in 1:nY) {
      #cdfvmat[i, ] <- apply((cqfv <= gridY[i]), 2, mean);
      cdfvmat[i, ] <- apply((cqfv <= gridY[i]), 2, sum);
      i0           <- which(cdfvmat[i, ]==0)
      if(length(i0)>0){
        cdfvmat[i, i0]  <- taus2[1];
        cdfvmat[i, -i0] <- taus2[cdfvmat[i, -i0]];
      }
      if(length(i0)==0){ cdfvmat[i, ] <- taus2[cdfvmat[i,]] }
    }
    #)
    #time2 <- system.time(
    #cdfvmat <- foreach(i=1:nY,.combine=rbind, .export = "cdfvmat_sub", .options.RNG = seed) %dorng% cdfvmat_sub(y0=gridY[i],cqfv=cqfv,taus2=taus2)
    #cdfvmat <- foreach(i=1:nY,.combine=rbind, .export = "cdfvmat_sub") %dopar% cdfvmat_sub(y0=gridY[i],cqfv=cqfv,taus2=taus2)
    #cdfvmat <- foreach(i=1:nY,.combine=rbind, .export = "cdfvmat_sub") %do% cdfvmat_sub(y0=gridY[i],cqfv=cqfv,taus2=taus2)
    #)
    #print(time1)
    #print(time2)
    
  }
  
  if(SS == "DR"){
    if(link == "logit"){  cdfvmat <- sapply(1:ncol(coefs), function(m) plogis(coefs[,m] %*% x_iv) ) }
    if(link == "probit"){ cdfvmat <- sapply(1:ncol(coefs), function(m) pnorm( coefs[,m] %*% x_iv) ) }
    cdfvmat <- t(cdfvmat)
  }
  
  return(cdfvmat)
}        


cdfvmat_sub <- function(y0,cqfv,taus2){
  
  cdfvmat     <- apply((cqfv <= y0), 2, sum);
  i0          <- which(cdfvmat==0)
  if(length(i0)>0){
    cdfvmat[i0]  <- taus2[1];
    cdfvmat[-i0] <- taus2[cdfvmat[-i0]];
  }
  if(length(i0)==0){ cdfvmat <- taus2[cdfvmat] }
  
  return(cdfvmat)
}


